@@ -20,6 +20,7 @@ import numpy as np
 
 from abc import abstractmethod
 from abc import ABC
+from abc import ABCMeta
 from tqdm import tqdm
 from typing import Any
 from typing import Set
@@ -32,6 +33,7 @@ from typing import TypeVar
 from typing import Callable
 from typing import Iterable
 from typing import Optional
+from typing import Protocol
 from typing import NamedTuple
 from argparse import Namespace
 from datetime import datetime
@@ -41,9 +43,12 @@ from functools import partial
 from itertools import product
 from collections import OrderedDict
 from dataclasses import asdict
+from dataclasses import fields
 from dataclasses import dataclass
+from dataclasses import Field
 
 from .types import configs_type
+from .types import np_dict_type
 from .types import general_config_type
 from .constants import TIME_FORMAT
 
@@ -54,6 +59,14 @@ dill._dill._reverse_typemap["ClassType"] = type
 # util functions
 
 
+TFnResponse = TypeVar("TFnResponse")
+
+
+class Fn(Protocol):
+    def __call__(self, *args: Any, **kwargs: Any) -> TFnResponse:
+        pass
+
+
 def walk(
     root: str,
     hierarchy_callback: Callable[[List[str], str], None],
@@ -92,6 +105,22 @@ def check_requires(fn: Any, name: str, strict: bool = True) -> bool:
     return False
 
 
+def get_requirements(fn: Any) -> List[str]:
+    if isinstance(fn, type):
+        fn = fn.__init__  # type: ignore
+    requirements = []
+    signature = inspect.signature(fn)
+    for k, param in signature.parameters.items():
+        if param.kind is inspect.Parameter.VAR_KEYWORD:
+            continue
+        if param.kind is inspect.Parameter.VAR_POSITIONAL:
+            continue
+        if param.default is not inspect.Parameter.empty:
+            continue
+        requirements.append(k)
+    return requirements
+
+
 def filter_kw(
     fn: Callable,
     kwargs: Dict[str, Any],
@@ -105,7 +134,7 @@ def filter_kw(
     return kw
 
 
-def safe_execute(fn: Any, kw: Dict[str, Any], *, strict: bool = False) -> Any:
+def safe_execute(fn: Fn, kw: Dict[str, Any], *, strict: bool = False) -> TFnResponse:
     return fn(**filter_kw(fn, kw, strict=strict))
 
 
@@ -122,11 +151,16 @@ def get_num_positional_args(fn: Callable) -> Union[int, float]:
     return counter
 
 
-def prepare_workplace_from(workplace: str, timeout: timedelta = timedelta(30)) -> str:
+def prepare_workspace_from(
+    workspace: str,
+    *,
+    timeout: timedelta = timedelta(30),
+    make: bool = True,
+) -> str:
     current_time = datetime.now()
-    if os.path.isdir(workplace):
-        for stuff in os.listdir(workplace):
-            if not os.path.isdir(os.path.join(workplace, stuff)):
+    if os.path.isdir(workspace):
+        for stuff in os.listdir(workspace):
+            if not os.path.isdir(os.path.join(workspace, stuff)):
                 continue
             try:
                 stuff_time = datetime.strptime(stuff, TIME_FORMAT)
@@ -134,29 +168,30 @@ def prepare_workplace_from(workplace: str, timeout: timedelta = timedelta(30)) -
                 if stuff_delta > timeout:
                     msg = f"{stuff} will be removed (already {stuff_delta} ago)"
                     print_warning(msg)
-                    shutil.rmtree(os.path.join(workplace, stuff))
+                    shutil.rmtree(os.path.join(workspace, stuff))
             except:
                 pass
-    workplace = os.path.join(workplace, current_time.strftime(TIME_FORMAT))
-    os.makedirs(workplace)
-    return workplace
+    workspace = os.path.join(workspace, current_time.strftime(TIME_FORMAT))
+    if make:
+        os.makedirs(workspace)
+    return workspace
 
 
-def get_latest_workplace(root: str) -> Optional[str]:
+def get_latest_workspace(root: str) -> Optional[str]:
     if not os.path.isdir(root):
         return None
-    all_workplaces = []
+    all_workspaces = []
     for stuff in os.listdir(root):
         if not os.path.isdir(os.path.join(root, stuff)):
             continue
         try:
             datetime.strptime(stuff, TIME_FORMAT)
-            all_workplaces.append(stuff)
+            all_workspaces.append(stuff)
         except:
             pass
-    if not all_workplaces:
+    if not all_workspaces:
         return None
-    return os.path.join(root, sorted(all_workplaces)[-1])
+    return os.path.join(root, sorted(all_workspaces)[-1])
 
 
 def sort_dict_by_value(d: Dict[Any, Any], *, reverse: bool = False) -> OrderedDict:
@@ -427,15 +462,50 @@ def check(constraints: Dict[str, Union[str, List[str]]], *, raise_error: bool =
 
 
 TRegister = TypeVar("TRegister", bound="WithRegister", covariant=True)
+TSerializable = TypeVar("TSerializable", bound="ISerializable", covariant=True)
+TSerializableArrays = TypeVar(
+    "TSerializableArrays",
+    bound="ISerializableArrays",
+    covariant=True,
+)
+TSArrays = TypeVar("TSArrays", bound="ISerializableArrays", covariant=True)
+TSDataClass = TypeVar("TSDataClass", bound="ISerializableDataClass", covariant=True)
 TDataClass = TypeVar("TDataClass", bound="DataClassBase")
 
+serializable_dataclasses: Dict[str, Type["ISerializableDataClass"]] = {}
+
+
+@dataclass
+class DataClassBase(ABC):
+    @property
+    def fields(self) -> List[Field]:
+        return fields(self)
+
+    @property
+    def field_names(self) -> List[str]:
+        return [f.name for f in self.fields]
+
+    @property
+    def attributes(self) -> List[Any]:
+        return [getattr(self, name) for name in self.field_names]
+
+    def asdict(self) -> Dict[str, Any]:
+        return asdict(self)
+
+    def copy(self: TDataClass) -> TDataClass:
+        return type(self)(**shallow_copy_dict(asdict(self)))
+
+    def update_with(self: TDataClass, other: TDataClass) -> TDataClass:
+        d = update_dict(other.asdict(), self.asdict())
+        return self.__class__(**d)
+
 
 class WithRegister(Generic[TRegister]):
     d: Dict[str, Type[TRegister]]
     __identifier__: str
 
     @classmethod
-    def get(cls, name: str) -> Type[TRegister]:
+    def get(cls: Type[TRegister], name: str) -> Type[TRegister]:
         return cls.d[name]
 
     @classmethod
@@ -443,24 +513,35 @@ class WithRegister(Generic[TRegister]):
         return name in cls.d
 
     @classmethod
-    def make(cls, name: str, config: Dict[str, Any]) -> TRegister:
-        return cls.get(name)(**config)  # type: ignore
+    def make(
+        cls: Type[TRegister],
+        name: str,
+        config: Dict[str, Any],
+        *,
+        ensure_safe: bool = False,
+    ) -> TRegister:
+        base = cls.get(name)
+        if not ensure_safe:
+            return base(**config)  # type: ignore
+        return safe_execute(base, config)
 
     @classmethod
     def make_multiple(
-        cls,
+        cls: Type[TRegister],
         names: Union[str, List[str]],
         configs: configs_type = None,
+        *,
+        ensure_safe: bool = False,
     ) -> List[TRegister]:
         if configs is None:
             configs = {}
         if isinstance(names, str):
             assert isinstance(configs, dict)
-            return cls.make(names, configs)  # type: ignore
+            return cls.make(names, configs, ensure_safe=ensure_safe)  # type: ignore
         if not isinstance(configs, list):
             configs = [configs.get(name, {}) for name in names]
         return [
-            cls.make(name, shallow_copy_dict(config))
+            cls.make(name, shallow_copy_dict(config), ensure_safe=ensure_safe)
             for name, config in zip(names, configs)
         ]
 
@@ -491,12 +572,251 @@ class WithRegister(Generic[TRegister]):
 
 
 @dataclass
-class DataClassBase(ABC):
-    def asdict(self) -> Dict[str, Any]:
-        return asdict(self)
+class JsonPack(DataClassBase):
+    type: str
+    info: Dict[str, Any]
 
-    def copy(self: TDataClass) -> TDataClass:
-        return type(self)(**shallow_copy_dict(asdict(self)))
+
+class ISerializable(WithRegister, Generic[TSerializable], metaclass=ABCMeta):
+    # abstract
+
+    @abstractmethod
+    def to_info(self) -> Dict[str, Any]:
+        pass
+
+    @abstractmethod
+    def from_info(self, info: Dict[str, Any]) -> None:
+        pass
+
+    # optional callbacks
+
+    def after_load(self) -> None:
+        pass
+
+    # api
+
+    def to_pack(self) -> JsonPack:
+        return JsonPack(self.__identifier__, self.to_info())
+
+    @classmethod
+    def from_pack(cls: Type[TSerializable], pack: Dict[str, Any]) -> TSerializable:
+        obj: ISerializable = cls.make(pack["type"], {})
+        obj.from_info(pack["info"])
+        obj.after_load()
+        return obj
+
+    def to_json(self) -> str:
+        return json.dumps(self.to_pack().asdict())
+
+    @classmethod
+    def from_json(cls: Type[TSerializable], json_string: str) -> TSerializable:
+        return cls.from_pack(json.loads(json_string))
+
+    def copy(self: TSerializable) -> TSerializable:
+        copied = self.__class__()
+        copied.from_info(shallow_copy_dict(self.to_info()))
+        return copied
+
+
+class PureFromInfoMixin:
+    def from_info(self, info: Dict[str, Any]) -> None:
+        for k, v in info.items():
+            setattr(self, k, v)
+
+
+class ISerializableArrays(ISerializable, Generic[TSArrays], metaclass=ABCMeta):
+    @abstractmethod
+    def to_npd(self) -> np_dict_type:
+        pass
+
+    @abstractmethod
+    def from_npd(self, npd: np_dict_type) -> None:
+        pass
+
+    def copy(self: TSerializableArrays) -> TSerializableArrays:
+        copied: TSerializableArrays = super().copy()
+        copied.from_npd(shallow_copy_dict(self.to_npd()))
+        return copied
+
+
+@dataclass
+class ISerializableDataClass(
+    ISerializable,
+    DataClassBase,
+    Generic[TSDataClass],
+    metaclass=ABCMeta,
+):
+    def to_info(self) -> Dict[str, Any]:
+        return self.asdict()
+
+    def from_info(self, info: Dict[str, Any]) -> None:
+        for k, v in info.items():
+            setattr(self, k, v)
+
+    @classmethod
+    def get(cls: Type[TRegister], name: str) -> Type[TRegister]:
+        return serializable_dataclasses[name]
+
+    @classmethod
+    def has(cls, name: str) -> bool:
+        return name in serializable_dataclasses
+
+    @classmethod
+    def make(cls: Type[TRegister], name: str, config: Dict[str, Any]) -> TRegister:
+        return cls.get(name)(**config)  # type: ignore
+
+    @classmethod
+    def register(
+        cls,
+        name: str,
+        *,
+        allow_duplicate: bool = False,
+    ) -> Callable:
+        def before(cls_: Type) -> None:
+            cls_.__identifier__ = name
+
+        return register_core(
+            name,
+            serializable_dataclasses,
+            allow_duplicate=allow_duplicate,
+            before_register=before,
+        )
+
+    @classmethod
+    def remove(cls, name: str) -> Callable:
+        return serializable_dataclasses.pop(name)
+
+    @classmethod
+    def check_subclass(cls, name: str) -> bool:
+        return issubclass(serializable_dataclasses[name], cls)
+
+
+class Serializer:
+    id_file: str = "id.txt"
+    info_file: str = "info.json"
+    npd_folder: str = "npd"
+
+    @classmethod
+    def save_info(
+        cls,
+        folder: str,
+        *,
+        info: Optional[Dict[str, Any]] = None,
+        serializable: Optional[ISerializable] = None,
+    ) -> None:
+        os.makedirs(folder, exist_ok=True)
+        if info is None and serializable is None:
+            raise ValueError("either `info` or `serializable` should be provided")
+        if info is None:
+            info = serializable.to_info()
+        with open(os.path.join(folder, cls.info_file), "w") as f:
+            json.dump(info, f)
+
+    @classmethod
+    def load_info(cls, folder: str) -> Dict[str, Any]:
+        return cls.try_load_info(folder, strict=True)
+
+    @classmethod
+    def try_load_info(
+        cls,
+        folder: str,
+        *,
+        strict: bool = False,
+    ) -> Optional[Dict[str, Any]]:
+        info_path = os.path.join(folder, cls.info_file)
+        if not os.path.isfile(info_path):
+            if not strict:
+                return
+            raise ValueError(f"'{info_path}' does not exist")
+        with open(info_path, "r") as f:
+            info = json.load(f)
+        return info
+
+    @classmethod
+    def save_npd(
+        cls,
+        folder: str,
+        *,
+        npd: Optional[np_dict_type] = None,
+        serializable: Optional[ISerializableArrays] = None,
+    ) -> None:
+        os.makedirs(folder, exist_ok=True)
+        if npd is None and serializable is None:
+            raise ValueError("either `npd` or `serializable` should be provided")
+        if npd is None:
+            npd = serializable.to_npd()
+        npd_folder = os.path.join(folder, cls.npd_folder)
+        os.makedirs(npd_folder, exist_ok=True)
+        for k, v in npd.items():
+            np.save(os.path.join(npd_folder, f"{k}.npy"), v)
+
+    @classmethod
+    def load_npd(cls, folder: str) -> np_dict_type:
+        os.makedirs(folder, exist_ok=True)
+        npd_folder = os.path.join(folder, cls.npd_folder)
+        if not os.path.isdir(npd_folder):
+            raise ValueError(f"'{npd_folder}' does not exist")
+        npd = {}
+        for file in os.listdir(npd_folder):
+            key = os.path.splitext(file)[0]
+            npd[key] = np.load(os.path.join(npd_folder, file))
+        return npd
+
+    @classmethod
+    def save(
+        cls,
+        folder: str,
+        serializable: ISerializable,
+        *,
+        save_npd: bool = True,
+    ) -> None:
+        cls.save_info(folder, serializable=serializable)
+        if save_npd and isinstance(serializable, ISerializableArrays):
+            cls.save_npd(folder, serializable=serializable)
+        with open(os.path.join(folder, cls.id_file), "w") as f:
+            f.write(serializable.__identifier__)
+
+    @classmethod
+    def load(
+        cls,
+        folder: str,
+        base: Type[TSerializable],
+        *,
+        swap_id: Optional[str] = None,
+        swap_info: Optional[Dict[str, Any]] = None,
+        load_npd: bool = True,
+    ) -> TSerializable:
+        serializable = cls.load_empty(folder, base, swap_id=swap_id)
+        serializable.from_info(swap_info or cls.load_info(folder))
+        if load_npd and isinstance(serializable, ISerializableArrays):
+            serializable.from_npd(cls.load_npd(folder))
+        return serializable
+
+    @classmethod
+    def load_empty(
+        cls,
+        folder: str,
+        base: Type[TSerializable],
+        *,
+        swap_id: Optional[str] = None,
+    ) -> TSerializable:
+        if swap_id is not None:
+            s_type = swap_id
+        else:
+            id_path = os.path.join(folder, cls.id_file)
+            if not os.path.isfile(id_path):
+                raise ValueError(f"cannot find '{id_path}'")
+            with open(id_path, "r") as f:
+                s_type = f.read().strip()
+        return base.make(s_type, {})
+
+
+class IWithRequirements:
+    @classmethod
+    def requirements(cls) -> List[str]:
+        requirements = get_requirements(cls)
+        requirements.remove("self")
+        return requirements
 
 
 class SanityChecker:
@@ -1812,12 +2132,12 @@ class lock_manager(context_error_handler, LoggingMixin):
     Examples
     --------
     >>> import dill
-    >>> workplace = "_cache"
+    >>> workspace = "_cache"
     >>> target_write_files_full_path = [
-    >>>     os.path.join(workplace, "file1.pkl"),
-    >>>     os.path.join(workplace, "file2.pkl")
+    >>>     os.path.join(workspace, "file1.pkl"),
+    >>>     os.path.join(workspace, "file2.pkl")
     >>> ]
-    >>> with lock_manager(workplace, target_write_files_full_path):
+    >>> with lock_manager(workspace, target_write_files_full_path):
     >>>     for write_file_full_path in target_write_files_full_path:
     >>>         with open(write_file_full_path, "wb") as wf:
     >>>             dill.dump(..., wf)
@@ -1829,7 +2149,7 @@ class lock_manager(context_error_handler, LoggingMixin):
 
     def __init__(
         self,
-        workplace,
+        workspace,
         stuffs,
         verbose_level=None,
         set_lock=True,
@@ -1837,10 +2157,10 @@ class lock_manager(context_error_handler, LoggingMixin):
         name=None,
         wait=1000,
     ):
-        self._workplace = workplace
+        self._workspace = workspace
         self._verbose_level = verbose_level
         self._name, self._wait = name, wait
-        os.makedirs(workplace, exist_ok=True)
+        os.makedirs(workspace, exist_ok=True)
         self._stuffs, self._set_lock = stuffs, set_lock
         self._clear_stuffs = clear_stuffs_after_exc
         self._is_locked = False
@@ -1887,7 +2207,7 @@ class lock_manager(context_error_handler, LoggingMixin):
                         f.write(
                             f"name      : {self._name}\n"
                             f"timestamp : {timestamp()}\n"
-                            f"workplace : {self._workplace}\n"
+                            f"workspace : {self._workspace}\n"
                             f"stuffs    :\n{self.cache_stuffs_str}"
                         )
                     self.__refresher = _lock_file_refresher(self.lock_file)
@@ -1985,7 +2305,7 @@ class lock_manager(context_error_handler, LoggingMixin):
 
     @property
     def lock_file(self):
-        return os.path.join(self._workplace, self.__lock__)
+        return os.path.join(self._workspace, self.__lock__)
 
     @property
     def logging_suffix(self):
